#!/usr/bin/env python

"""
This script is a very simple wrapper around three unix
programs: cdparanoia (http://www.xiph.org/paranoia/),
sox (http://home.sprynet.com/~cbagwell/sox.html) and oggenc
(http://www.vorbis.com/).  The purpose of this script is to rip audio
tracks, normalize their sound and compress them to .ogg format, all of
that automatically.

The process is simple: for each track, cdparanoia is called to extract
the audio to a temporary wav file. Sox is used on it to see if it is
normalized.  If it is, oggenc is called on it to generate the apropiate
.ogg version.  If it isn't normalized, sox is called on it to normalize
it, and the output is fed directly to oggenc. The result is that you only
need enough hard disk space to hold the track being compressed, and the
generated ogg files. It could be made to not store _any_ temporary file,
but in the case of tracks needed to be normalized you would be reading
the track twice, which at least for me is the speed bottleneck.

This script is gift-ware, it's given to you freely as a gift.  You may
use, modify, redistribute, and generally hack it about in any way you
like, and you do not have to give me anything in return. However, if
you like this script you are encouraged to thank me emailing me your
opinion about it, requests for more features, or directly diffs which
implement them. If you redistribute parts of this script or use it
somewhere successfully, it would be nice if you mentioned me somewhere
in the credits, but you are not required to do this.

Of course, I do not accept responsibility for any effects, adverse or
otherwise, that this code may have on you, your computer, your sanity,
your dog and anything else that you can think of.

Use it at your own risk.

Written by Grzegorz Adam Hankiewicz. Possible contact emails:
gradha@terra.es, gradha@users.sourceforge.net, gradha@iname.com.
Download this and other things from http://gradha.infierno.org/.
"""

import os, popen2, string, re, sys

BUFSIZE = 1024*100
version = "0.2 ogg-cdparanoia.py"
bitrate = 0

def run_command(command):
   """Runs the specified command and waits for the output. Abnormal
   termination will terminate the execution flow."""
   pid = os.fork()

   if pid == 0:
      os.execvp (command[0], command)
      sys.exit (0) # por si las mosquis

   proc, ret = os.waitpid (pid, 0)
   if ret:
      print "Valor de retorno anmalo: %d" % ret
      sys.exit (1)



def read_command_output (command):
   """Similar to run_command, but the output produced by the command is
   returned as a list."""
   child_stdin, child_stdout, child_stderr = popen2.popen3(command)
   list = []
   try:
      for f in map(string.strip, child_stdout.readlines()): list.append(f)
   except: pass
   try:
      for f in map(string.strip, child_stderr.readlines()): list.append(f)
   except: pass
   child_stdin.close()
   child_stdout.close()
   child_stderr.close()
   return list

   

def get_track_list():
   """Queries cdparanoia about available audio tracks, and returns a list of
   integers with those tracks."""
   list = []
   exp = re.compile("\d+\.")

   for f in read_command_output(["cdparanoia", "-Q"]):
      res = exp.match(f)
      if res:
         list.append(int(string.split(f, ".")[0]))

   return list



def data_reader(pipe, file, volume):
   """Reads the temporary wav file through sox normalizing it's sound. The
   result is piped to the data writer."""
   os.close(pipe[0])
   pipe = pipe[1]

   try:
      child_stdin, child_stdout, child_stderr = popen2.popen3(["sox", "-v",
         "%f" % volume, "-t", "wav", file, "-t", "wav", "-"])
      data = child_stdin.read(BUFSIZE)
      while data:
         bytes = os.write(pipe, data)
         if bytes != len(data):
            raise IOError, "[READER]: Couldn't write data to pipe! (%d != %d)" % (bytes, len(data))
         data = child_stdin.read(BUFSIZE)

      os.close(pipe)
      child_stdin.close()
      child_stdout.close()
      child_stderr.close()

   except:
      os.close(pipe)
      raise

   sys.exit(0)



def data_writer(pipe, track, son_pid):
   """Expects data to flow through the pipe, which will feed directly to
   oggenc, generating the apropiate ogg file."""
   os.close(pipe[1])
   pipe = pipe[0]

   command = ["oggenc", "-Q", "-", "-o", "track%02d.ogg" % track]
   if bitrate:
      command.append("-b")
      command.append("%d" % bitrate)
   child_stdin, child_stdout, child_stderr = popen2.popen3(command)

   data = os.read(pipe, BUFSIZE)
   while data:
      child_stdout.write(data)
      data = os.read(pipe, BUFSIZE)

   os.close(pipe)
   child_stderr.close()
   child_stdout.close()
   child_stdin.close()

   proc, ret = os.waitpid(son_pid, 0)
   if ret:
      raise "[WRITER]: Error waiting for reader to end (return code %d)" % ret

      

def get_volume(file):
   """When run over a wav file, returns the tuple (doit, volume). volume
   is a float value containing the volume to which the file should be
   normalized. doit is a boolean value which gives a recommendation to
   what to do: normalize or not the file."""
   out = read_command_output(["sox", "-t", "wav", file, "-e", "stat", "-v"])
   volume = float (out[0])

   if volume < 1.1:
      return 0, volume
      
   return 1, volume



def rip_festival():
   """Function which controls the ripping."""
   list = get_track_list()
   print "Available audio tracks: ", list
   for f in list:
      print "Ripping '%d'" % f
      run_command(["cdparanoia", "-q", "%d" % f, "temp.wav"])
      need_amplification, volume = get_volume("temp.wav")
      if need_amplification:
         pipe = os.pipe()
         parent = os.fork()
         if parent:  data_writer(pipe, f, parent)
         else:       data_reader(pipe, "temp.wav", volume)
      else:
         command = ["oggenc", "-Q", "temp.wav", "-o", "track%02d.ogg" % f]
         if bitrate:
            command.append("-b")
            command.append("%d" % bitrate)
         run_command(command)
         
      os.unlink("temp.wav")



def show_help():
   """Show commandline help."""
   print "ogg-cdparanoia.py written by Grzegorz Adam Hankiewicz <gradha@terra.es>"
   print "This script is gift-ware."
   print
   print " -h       Show this help"
   print " -v       Show version number and exit"
   print " -rip     Rip all audio tracks to .ogg files"
   print " -64      Encode at average of 64 Kbps"

   

if __name__ == "__main__":   
   """Entry point for userland."""
   params = sys.argv[1:]
   if not len(params) or "-h" in params:
      show_help()
      sys.exit(0)
      
   if "-v" in params:
      print version
      sys.exit(0)
      
   if "-64" in params:
      print "Encoding at 64 Kbps."
      bitrate = 64
      
   if "-rip" in params:
      rip_festival()
      sys.exit(0)

   show_help()
   
